#include <csignal>
#include <cstring>
#include <cstdio>
#include <cctype>
#include <algorithm>
#include <unistd.h>

#include <getopt.h>

#include <string>
#include <vector>
#include <cerrno>

using namespace std;

#include "serverfun.h"
#include "resample.hh"
#include "filters.h"

static unsigned Filter = 1;  /* linear */
static double blur = 1;      /* <1 = sharp, >1 = blurry */
static bool dbl1=false, dbl2=false, flip1=false, flip2=false;

static void CheckFlip(stent *Buf, unsigned count)
{
   	if(dbl1) for(unsigned a=0; a<count; ++a) Buf[a].l += Buf[a].l;
   	if(dbl2) for(unsigned a=0; a<count; ++a) Buf[a].r += Buf[a].r;
   	if(flip1) for(unsigned a=0; a<count; ++a) Buf[a].l = -Buf[a].l;
   	if(flip2) for(unsigned a=0; a<count; ++a) Buf[a].r = -Buf[a].r;
}

struct ContiguousResampleBuf
{ 
public:
    double center;
public:
    ContiguousResampleBuf() : center(0.5), firstpos(0) { }
    
    inline void NotNeedOlderThan(int pos)
    {
        const int n_discard_at_once = 4096;
        int discardable = pos-firstpos;
        if(discardable >= n_discard_at_once)
        {
            int n_discard = (discardable / n_discard_at_once) * n_discard_at_once;
            AudioData.erase(AudioData.begin(), AudioData.begin()+n_discard);
            firstpos += n_discard;
        }
    }
    inline bool AddMul(double& l, double& r, int pos, double coeff) const
    {
        if(pos >= firstpos)
        {
            if(pos >= (int)(firstpos + AudioData.size()))
                return false;
            l += AudioData[pos - firstpos].l * coeff;
            r += AudioData[pos - firstpos].r * coeff;
        }
        return true;
    }
    inline void Feed(const stent* input, unsigned nelems)
    {
        AudioData.insert(AudioData.end(), input, input+nelems);
    }
    inline int DropExcess()
    {
        //printf("excess=%d, center=%g\n", firstpos, center);
        int res = firstpos;
        firstpos = 0;
        return res;  
    }
    inline unsigned GetCapacity() const { return AudioData.size(); }
private:
    std::vector<stent> AudioData;
    int firstpos;
};

/* This the function which started the program development.
 * The core of this program. The rest is just waveform
 * conversions and parameter handling...
 */
static void Interpolate(Stream &input, Stream &output)
{
    if(input.rate() == output.rate())
    {
        // No change in rate, just copy
        stent tab[512];
        for(;;)
        {
            unsigned oco = input.Read(tab, 512);
            if(!oco)return;
            output.Write(tab, oco);
        }
    }

    const double factor = output.rate() / (double)input.rate();
    const double factor_rev = 1.0 / factor;
    
    double scale   = min(factor, 1.0) / blur;
    double support = Filters[Filter].support / scale;
    if(support <= 0.5) { support = 0.5 + 1E-12; scale = 1.0; }
    
    double *contribution = (double*)alloca((sizeof(double)) * ((unsigned)(support*2 + 2)));
    double (*const filterfun)(double) = Filters[Filter].function;

    ContiguousResampleBuf status;
    
    bool NoData = false;
    while(!NoData)
    {
        /* Refill input */
        {
         stent inbuf[4096];
         unsigned nread = input.Read(inbuf, sizeof(inbuf)/sizeof(*inbuf));
         status.Feed(inbuf, nread);
         if(nread <= 0) NoData = true;
        }
    TryConvertMore:
        std::vector<stent> result;
        result.reserve((unsigned)(status.GetCapacity() * factor) + 8);
        while(result.size() < 256)
        {
            const int start = (int)(status.center-support+0.5);
            const int stop  = (int)(status.center+support+0.5);
            double density = 0.0;
            
            unsigned nmax = stop-start;
            
            double s = start - status.center + 0.5;
            for(unsigned n=0; n<nmax; ++n, ++s)
            {
                contribution[n] = filterfun(s * scale);
                density += contribution[n];
            }
            if(density != 0.0 && density != 1.0)
            {
                for(unsigned n=0; n<nmax; ++n)
                    contribution[n] /= density;
            }
            
            double result_l = 0.0;
            double result_r = 0.0;
            for(unsigned n=0; n<nmax; ++n)
            {
                if(!status.AddMul(result_l, result_r, start+n, contribution[n]))
                {
                    status.NotNeedOlderThan(start);
                    goto NoMoreData;
                }
            }
            
            status.center += factor_rev;
            
            int res_l = (int)(result_l+0.5);
            if(res_l<-32768)res_l=-32768; if(res_l>32767)res_l=32767;
            
            int res_r = (int)(result_r+0.5);
            if(res_r<-32768)res_r=-32768; if(res_r>32767)res_r=32767;
            
            stent o; o.l = res_l; o.r = res_r;
            result.push_back(o);
        }
        output.Write(&result[0], result.size());
        goto TryConvertMore;
    NoMoreData:
        status.center -= status.DropExcess();
        output.Write(&result[0], result.size());
        output.Flush();
    }
}

Stream::Stream(const char *settings) : bit16(true), stereo(true), herz(0)
{
    while(*settings)
    {
        if(*settings=='m')stereo = false;
        else if(*settings=='s')stereo = true;
        else if(*settings=='b')bit16 = false;
        else if(*settings=='w')bit16 = true;
        else if(*settings=='r')herz = 0;
        else if(*settings>='0' && *settings<='9')
            herz = herz * 10 + (*settings - '0');
        else if(*settings!='-')throw *settings;
        ++settings;
    }
    if(!herz)herz = 44100;
}

unsigned Stream::Read(stent *dest, unsigned count) const
{
    int retval;
    if(bit16)
        if(stereo)
            retval = SafeRead((uint8*)dest, count*4, 4) / 4;
        else
            retval = read16(dest, count);
    else
        if(stereo)
            retval = readste(dest, count);
        else
            retval = read8(dest, count);
    if(retval > 0) CheckFlip(dest, retval);
	return retval;
}
void Stream::Write(const stent *dest, unsigned count) const
{
    if(bit16)
        if(stereo)
            SafeWrite(dest, count*4);
        else
            write16(dest, count);
    else
        if(stereo)
            writeste(dest, count);
        else
            write8(dest, count);
}

void Stream::Flush() const
{
    //fdatasync(1);
}

/* Read 16-bit mono entries */
unsigned Stream::read16(stent *dest, unsigned count)
{
    unsigned bytes = count * sizeof(int16);
    int16 *tab = (int16 *)alloca(bytes);
    count = SafeRead((uint8*)tab, bytes, 2) / 2;
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a];
    }
    return count;
}
/* Read 8-bit stereo entries */
unsigned Stream::readste(stent *dest, unsigned count)
{
    unsigned bytes = count * 2 * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    count = SafeRead((uint8*)tab, bytes, 2) / 2;
    for(unsigned a=0, b=0; b<count; ++b)
    {
        dest[b].r = tab[a++]*256 - 32768;
        dest[b].l = tab[a++]*256 - 32768;
    }
    return count;
}
/* Read 8-bit mono entries */
unsigned Stream::read8(stent *dest, unsigned count)
{
    unsigned bytes = count * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    count = SafeRead((uint8*)tab, bytes, 1);
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a]*256 - 32768;
    }
    return count;
}

void Stream::write16(const stent *src, unsigned count)
{
    unsigned bytes = count * sizeof(int16);
    int16 *tab = (int16 *)alloca(bytes);
    for(unsigned a=0; a<count; ++a)
    {
        /* 16-bit stereo to mono mixing */
        tab[a] = (src[a].r + src[a].l) / 2;
    }
    SafeWrite(tab, bytes);
}
void Stream::writeste(const stent *src, unsigned count)
{
    unsigned bytes = count * 2 * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    for(unsigned a=0, b=0; b<count; ++b)
    {
        tab[a++] = (src[b].r + 32768) / 256;
        tab[a++] = (src[b].l + 32768) / 256;
    }
    SafeWrite(tab, bytes);
    
}
void Stream::write8(const stent *src, unsigned count)
{
    unsigned bytes = count * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    for(unsigned a=0; a<count; ++a)
    {
        /* 8-bit stereo to mono mixing */
        tab[a] = ((src[a].r + src[a].l) / 2 + 32768) / 256;
    }
    SafeWrite(tab, bytes);
}

/* SafeRead
 *   Synopsis:
 *     -Try to read _some_ data from stdin
 *     -At most "nbytes" bytes
 *     -The amount of bytes read _must_ be aligned properly as indicated
 *     -If a misaligned amount is read, it will be buffered for later calls
 *     -Partial reads are ok.
 */
int Stream::SafeRead(uint8* target, unsigned nbytes, unsigned align)
{
    static std::vector<unsigned char> FractionalBytes;
TryMore:
    unsigned offs = 0;
    std::memcpy(target, &FractionalBytes[0], FractionalBytes.size());
    offs += FractionalBytes.size();
    
    unsigned try_read = ((nbytes - offs) / align) * align;
    
Retry:
    int res = read(0, target+offs, try_read);
    if(res == -1 && errno == EINTR) goto Retry;
    
    unsigned bytes_got    = offs + (res < 0 ? 0 : res);
    
    unsigned bytes_good   = bytes_got / align * align;
    unsigned bytes_excess = bytes_got - bytes_good;
    
    FractionalBytes.assign(target+bytes_good, target+bytes_got);
    
    if(res > 0 && bytes_good == 0) goto TryMore;
    return bytes_good;
}

#define VERSIONSTR \
    "resample v"VERSION" - Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "Portions copyright (C) 2001 ImageMagick Studio.\n" \
    "This program is under GPL. streamixer-"VERSION".tar.{gz,bz2}\n" \
    "are available at the homepage of the author.\n"

static void Killer(int)
{
    exit(-1);
}

int main(int argc, char** argv)
{
    signal(SIGPIPE, Killer);

    Stream *input = NULL, *output = NULL;
    bool Heelp = false;
    bool stderr_helper = false;
    
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",     0, 0,'h'},
            {"version",  0, 0,'V'},
            {"blur",     1, 0,'b'},
            {"lflip",    0, 0,'L'},
            {"rflip",    0, 0,'R'},
            {"ldouble",  0, 0,502},
            {"rdouble",  0, 0,503},
            {"filter",   1, 0,'f'},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "HhVb:s:f:", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'H': stderr_helper = true; break;
            case 'h': Heelp = true; break;
            case 'V':
                printf("%s", VERSIONSTR);
                return 0;
            case 'b': blur = 1 + strtol(optarg,NULL,10); break;
            case 's': blur = 1 - strtol(optarg,NULL,10); break;
            case 'L': flip1 = true; break;
            case 'R': flip2 = true; break;
            case 502: dbl1 = true; break;
            case 503: dbl2 = true; break;
            case 'f':
            {
                std::string s = optarg;
                unsigned a;
                for(a=0; a<FilterCount; ++a)
                    if(Filters[a].name == s)
                    {
                        Filter = a;
                        break;
                    }
                if(a < FilterCount)break;
                fprintf(stderr, "'%s' is not a supported filter.\n", s.c_str());
                return -1;
            }
            default:
                break;
        }
    }
    
    if(argc != optind+2)
    {
        Heelp = true;
    }
    
    if(Heelp)
    {
        fprintf(stderr,
            "%s\n"
            "Usage: %s <inputspec> <outputspec> [<options> ...]\n"
            "\n"
            "Reads stdin, writes stdout.\n"
            "\n"
            "input- and outputspec:\n"
            "   b=8-bit, m=mono, rxxx=xxx Hz, examples:\n"
          "       \"\"      = 16-bit 44100 Hz stereo\n"
            "       m       = 16-bit 44100 Hz mono\n"
            "       b       = 8-bit 44100 Hz stereo\n"
            "       mb      = 8-bit 44100 Hz mono\n"
            "       mr22050 = 16-bit 22050 Hz mono\n"
            "       br22050 = 8-bit 22050 Hz stereo\n"
            "   etc. '-' must not be used.\n"
            "\nOptions:\n"
            "  --help, -h              Help\n"
            "  --version, -V           Version information\n"
            "  --blur, -b <amount>     Blur: 0=normal, >0=blur (cpu intensive)\n"
            "  --lflip, -L             Flip left channel\n"
            "  --rflip, -R             Flip right channel\n"
            "  --ldouble               Multiply left channel amplitude by 2\n"
            "  --rdouble               Multiply right channel amplitude by 2\n"
            "  --sharpen, -s <amount>  Equivalent to giving negative number to --blur\n"
            "  --filter, -f <filter>   Select filter (interpolation scheme)\n"
            "\n"
            "Available filters:\n"
            "  none, linear, triangle, hermite, hanning, hamming,\n"
            "  blackman, gaussian, quadratic, cubic, catrom, mitchell,\n"
            "  lanczos, bessel, sinc\n"
            "Recommended filters: linear, lanczos, none\n"
            "\n",
            VERSIONSTR, argv[0]);

        return 0;
    }

    try
    {
        input = new Stream(argv[optind+0]);
        output = new Stream(argv[optind+1]);
    }
    catch(char c)
    {
        fprintf(stderr, "Invalid option: '%c' (-switches must be specified after stream parameters)\n", c);
        return -1;
    }
    
        
    if(stderr_helper)
    {
        static const char c = '!';
        write(2, &c, 1);
    }
    Interpolate(*input, *output);
    return 0;
}
